from redun.executors.aws_batch import AWSBatchExecutor  # noqa: F401
from redun.executors.aws_glue import AWSGlueExecutor  # noqa: F401
from redun.executors.local import LocalExecutor  # noqa: F401
from redun.file import Dir, File, ShardedS3Dataset  # noqa: F401
from redun.handle import Handle  # noqa: F401
from redun.namespace import get_current_namespace, namespace  # noqa: F401
from redun.scheduler import (  # noqa: F401
    Scheduler,
    apply_tags,
    catch,
    cond,
    get_current_scheduler,
    merge_handles,
    set_current_scheduler,
    throw,
)
from redun.scripting import script  # noqa: F401
from redun.task import PartialTask, Task, get_task_registry, task  # noqa: F401

__version__ = "0.8.6"



import pytest

from redun import Scheduler, task
from redun.expression import SimpleExpression, TaskExpression, ValueExpression
from redun.functools import identity


def test_value_expression(scheduler: Scheduler) -> None:
    """
    ValueExpression should evaluate to itself.
    """
    # Test evaluation.
    expr = ValueExpression(10)
    assert scheduler.run(expr) == 10

    # Test hashing.
    assert expr.get_hash() == "8ce63809f39a33d99358ccb2b18c8761d6764aea"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "value_type": "builtins.int",
        "value": b"\x80\x03K\n.",
    }

    # Test deserialization.
    expr2 = ValueExpression.__new__(ValueExpression)
    expr2.__setstate__(state)
    assert expr2.value == 10

    # Test valid checking.
    assert expr.is_valid()


def test_simple_expression(scheduler: Scheduler) -> None:
    # Test evaluation.
    expr: SimpleExpression[str] = SimpleExpression("getitem", (["a", "b", "c"], 1))
    assert scheduler.run(expr) == "b"

    expr = SimpleExpression("getitem", (["a", "b", "c"], 5))
    with pytest.raises(IndexError):
        scheduler.run(expr)

    expr = SimpleExpression("bad_func", (10,))
    with pytest.raises(NotImplementedError):
        scheduler.run(expr)

    # Test hashing.
    expr = SimpleExpression("getitem", (["a", "b", "c"], 1))
    assert expr.get_hash() == "52d8a5c4b3672fc3dc7d53d5eb7a675a51f62894"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "func_name": "getitem",
        "args": (
            b"\x80\x03]q\x00(X\x01\x00\x00\x00aq\x01X\x01\x00\x00\x00bq\x02X\x01"
            b"\x00\x00\x00cq\x03eK\x01\x86q\x04."
        ),
        "kwargs": b"\x80\x03}q\x00.",
    }

    expr2 = SimpleExpression.__new__(SimpleExpression)
    expr2.__setstate__(state)
    assert expr2.func_name == "getitem"
    assert expr2._upstreams == [(["a", "b", "c"], 1), {}]
    assert scheduler.run(expr2) == "b"

    # Test valid checking.
    assert expr2.is_valid()


class Inner:
    def __init__(self, a: int):
        self.a = a


def test_lazy_operators(scheduler: Scheduler) -> None:
    @task()
    def task1(x: int, y: int = 1) -> int:
        return x + y

    @task()
    def task2(x: str, y: int = 3) -> str:
        return x * y

    @task()
    def get_task():
        return task1

    true_expressions = [
        task1(0) == ValueExpression(1),
        task1(1) != ValueExpression(1),
        task1(2) < task1(3),
        task1(2) <= task1(3),
        task1(3) > task1(2),
        task1(3) >= task1(2),
        task1(4) + task1(5) == 11,
        task1(3) - task1(1) == 2,
        task1(1) * task1(2) == 6,
        task1(3) / task1(1) == 2,
        task2("a") + task2("bc") == "aaabcbcbc",
        identity(True) & identity(True),
        identity(False) | identity(True),
        get_task()(7) == 8,
        ValueExpression(Inner(7)).a == 7,
        ValueExpression(["a", "b", "c"])[1] == "b",
    ]

    for expr in true_expressions:
        assert isinstance(expr, SimpleExpression)
        assert scheduler.run(expr) is True

    false_expressions = [
        task1(0) != ValueExpression(1),
        task1(1) == ValueExpression(1),
        task1(2) >= task1(3),
        task1(2) > task1(3),
        task1(3) <= task1(2),
        task1(3) < task1(2),
        task1(4) + task1(5) != 11,
        task1(3) - task1(1) != 2,
        task1(1) * task1(2) != 6,
        task1(3) / task1(1) != 2,
        task2("a") + task2("bc") == "aaabcbcbcd",
        identity(False) & identity(True),
        identity(False) | (False),
        get_task()(7) == 9,
        ValueExpression(Inner(7)).a == 8,
        ValueExpression(["a", "b", "c"])[1] == "f",
    ]

    for expr in false_expressions:
        assert isinstance(expr, SimpleExpression)
        assert scheduler.run(expr) is False


def test_task_expression(scheduler: Scheduler) -> None:
    @task()
    def task1(x, y=1):
        return x + y

    expr = task1(10)
    assert isinstance(expr, TaskExpression)
    assert expr.task_name == "task1"
    assert expr.args == (10,)
    assert expr.kwargs == {}
    assert scheduler.run(expr) == 11

    expr = task1(10, y=3)
    assert expr.kwargs == {"y": 3}
    assert scheduler.run(expr) == 13

    @task()
    def buggy():
        raise ValueError("bugs")

    with pytest.raises(ValueError):
        scheduler.run(buggy())

    # Test hashing.
    expr = TaskExpression("task1", (10,), {})
    assert expr.get_hash() == "7076d9536c59c05862286df0d767468098a4f0ed"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "task_name": "task1",
        "args": b"\x80\x03K\n\x85q\x00.",
        "kwargs": b"\x80\x03}q\x00.",
        "task_options": {},
    }

    expr2 = TaskExpression.__new__(TaskExpression)
    expr2.__setstate__(state)
    assert expr2.task_name == "task1"
    assert scheduler.run(expr2) == 11

    # Test valid checking.
    assert expr2.is_valid()
    expr2.task_name = "bad_task"
    assert not expr2.is_valid()

    # Test call expressions.
    assert scheduler.run(SimpleExpression("call", (task1, (2,), {"y": 3}))) == 5


def test_ban_bool_coerce() -> None:
    """
    Coercision of an Expression to bool should be ban to avoid common bugs.
    """
    expr = ValueExpression(10)
    with pytest.raises(TypeError):
        if expr:
            pass


def test_ban_iter() -> None:
    """
    Ban iteration on expression to prevent common misuse.
    """

    @task()
    def task1():
        return [1, 2, 3]

    expr = task1()

    with pytest.raises(TypeError):
        for x in expr:
            pass

    with pytest.raises(TypeError):
        _ = iter(expr)



import glob
import os
import pickle
import shlex
import tarfile
import tempfile
import threading
import zipfile
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, Union

import boto3

from redun.file import File
from redun.hashing import hash_stream
from redun.scheduler import Job

# Constants.
REDUN_PROG = "redun"
REDUN_REQUIRED_VERSION = ">=0.4.1"
DEFAULT_AWS_REGION = "us-west-2"

# S3 scratch filenames.
S3_SCRATCH_INPUT = "input"
S3_SCRATCH_OUTPUT = "output"
S3_SCRATCH_CODE = "code.tar.gz"
S3_SCRATCH_ERROR = "error"
S3_SCRATCH_HASHES = "eval_hashes"
S3_SCRATCH_STATUS = "status"

# Cache for AWS Clients.
_boto_clients: Dict[Tuple[int, str, str], boto3.Session] = {}


class JobStatus(NamedTuple):
    all: List[str]
    pending: List[str]
    inflight: List[str]
    success: List[str]
    failure: List[str]
    stopped: List[str]
    timeout: List[str]


def get_aws_client(service: str, aws_region: str = DEFAULT_AWS_REGION) -> boto3.Session:
    """
    Get an AWS Client with caching.
    """
    cache_key = (threading.get_ident(), service, aws_region)
    client = _boto_clients.get(cache_key)
    if not client:
        # Boto is not thread safe, so we create a client per thread using
        # `threading.get_ident()` as part of our cache key.
        # We need to create the client using a session to avoid all clients
        # sharing the same global session.
        # See: https://github.com/boto/boto3/issues/801#issuecomment-440942144
        session = boto3.session.Session()
        client = _boto_clients[cache_key] = session.client(service, region_name=aws_region)

    return client


def find_code_files(
    basedir: str = ".", includes: Optional[List[str]] = None, excludes: Optional[List[str]] = None
) -> Iterable[str]:
    """
    Find all the workflow code files consistent with the include/exclude patterns.
    """
    if includes is None:
        includes = ["**/*.py"]
    if excludes is None:
        excludes = []

    files: Set[str] = set()
    for pattern in includes:
        files |= set(glob.glob(os.path.join(basedir, pattern), recursive=True))

    for pattern in excludes:
        files -= set(glob.glob(os.path.join(basedir, pattern), recursive=True))
    return files


def create_tar(tar_path: str, file_paths: Iterable[str]) -> File:
    """
    Create a tar file from local file paths.
    """
    tar_file = File(tar_path)

    with tar_file.open("wb") as out:
        with tarfile.open(fileobj=out, mode="w|gz") as tar:
            for file_path in file_paths:
                tar.add(file_path)

    return tar_file


def extract_tar(tar_file: File, dest_dir: str = ".") -> None:
    """
    Extract a tar file to local paths.
    """
    with tar_file.open("rb") as infile:
        with tarfile.open(fileobj=infile, mode="r|gz") as tar:
            tar.extractall(dest_dir)


def create_zip(zip_path: str, base_path: str, file_paths: Iterable[str]) -> File:
    """
    Create a zip file from local file paths.
    """
    zip_file = File(zip_path)

    with zip_file.open("wb") as out:
        with zipfile.ZipFile(out, mode="w") as stream:
            for file_path in file_paths:
                arcname = os.path.relpath(file_path, base_path)
                stream.write(file_path, arcname)

    return zip_file


def get_job_scratch_dir(s3_scratch_prefix: str, job: Job) -> str:
    """
    Returns s3 scratch directory for a redun Job.
    """
    assert job.eval_hash
    return os.path.join(s3_scratch_prefix, "jobs", job.eval_hash)


def get_job_scratch_file(s3_scratch_prefix: str, job: Job, filename: str) -> str:
    """
    Returns s3 scratch path for a file related to a redun Job.
    """
    assert job.eval_hash
    return os.path.join(s3_scratch_prefix, "jobs", job.eval_hash, filename)


def get_code_scratch_file(s3_scratch_prefix: str, tar_hash: str, use_zip: bool = False) -> str:
    """
    Returns s3 scratch path for a code package tar file.
    """
    return os.path.join(s3_scratch_prefix, "code", tar_hash + (".zip" if use_zip else ".tar.gz"))


def get_array_scratch_file(s3_scratch_prefix: str, job_array_id: str, filename: str) -> str:
    """
    Returns an S3 scratch path for a file related to an AWS batch job
    """
    return os.path.join(s3_scratch_prefix, "array_jobs", job_array_id, filename)


def copy_to_s3(file_path: str, s3_scratch_dir: str) -> str:
    """
    Copies a file to the S3 scratch directory if it is not already on S3.
    Returns the path to the file on S3.
    """
    file = File(file_path)
    _, filename = os.path.split(file.path)

    s3_temp_file = File(f"{s3_scratch_dir.rstrip('/')}/{filename}")
    file.copy_to(s3_temp_file)
    return s3_temp_file.path


def get_default_region() -> str:
    """
    Returns the default AWS region.
    """
    return boto3.Session().region_name or DEFAULT_AWS_REGION


def get_aws_user(aws_region: str = DEFAULT_AWS_REGION) -> str:
    """
    Returns the current AWS user.
    """
    sts_client = get_aws_client("sts", aws_region=aws_region)
    response = sts_client.get_caller_identity()
    return response["Arn"]


def parse_code_package_config(config) -> Union[dict, bool]:
    """
    Parse the code package options from a AWSBatchExecutor config.
    """
    if not config.getboolean("code_package", fallback=True):
        return False

    include_config = config.get("code_includes", "**/*.py")
    exclude_config = config.get("code_excludes", "")

    return {"includes": shlex.split(include_config), "excludes": shlex.split(exclude_config)}


def package_code(s3_scratch_prefix: str, code_package: dict = {}, use_zip=False) -> File:
    """
    Package code to S3.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        file_paths = find_code_files(
            includes=code_package.get("includes"), excludes=code_package.get("excludes")
        )

        if use_zip:
            temp_file = File(os.path.join(tmpdir, "code.zip"))
            create_zip(temp_file.path, ".", file_paths)
        else:
            temp_file = File(os.path.join(tmpdir, "code.tar.gz"))
            create_tar(temp_file.path, file_paths)

        with temp_file.open("rb") as infile:
            tar_hash = hash_stream(infile)
        code_file = File(get_code_scratch_file(s3_scratch_prefix, tar_hash, use_zip=use_zip))
        if not code_file.exists():
            temp_file.copy_to(code_file)

    return code_file


def parse_task_result(s3_scratch_prefix: str, job: Job) -> Any:
    """
    Parse task result from s3 scratch path.
    """
    output_path = get_job_scratch_file(s3_scratch_prefix, job, S3_SCRATCH_OUTPUT)

    output_file = File(output_path)
    assert job.task
    if not job.task.script:
        with output_file.open("rb") as infile:
            result = pickle.load(infile)
    else:
        result = [0, output_file.read(mode="rb")]  # TODO: Get real exitcode.

    return result
